open Big_int
open Turing
open Table
open Macro_strip
open Macro_machine

module type TFind_BB =
sig
  open Table

  val gen_tables : ttable -> int -> int -> tcell -> tstate ->
    (ttable * int * int) list
  val loop : int -> int -> ttable * tcount * tcount
end

module Find_BB =
struct
  open Table
  open Macro_Machine
  open Macro_Strip
    
  type semi_table = ttable * int * int
      
  let gen_tables table max_state numundef read state =
    if numundef = 1
    then
      let table' = Table.copy table in
	Table.add_halt table' read state;
	[(table',max_state,0)]
    else
      let l = ref []
      and n = min (max_state+1) ((Array.length table) - 1)
      in
	for i = 0 to 1 do
	  for j = 0 to n do
	    let t' = Table.copy table
	    and t'' = Table.copy table
	    and tr' = {write = i; move = Goleft; new_state = j}
	    and tr'' = {write = i; move = Goright; new_state = j}
	    and ms' = if j < n then max_state else n
	    and numundef' = numundef - 1
	    in
	      Table.add_trans t' read state tr';
	      Table.add_trans t'' read state tr'';
	      l := (t',ms',numundef')::(t'',ms',numundef')::(!l)
	  done
	done;
	!l
	  
  let loop n cell_size =
    let rec loop_aux champion score steps = function
      | [] -> champion,score,steps
      | (table,maxst,nundef)::q ->
	  begin
	    match Macro_Machine.run table cell_size with
	      | MACSTOP(str,steps') ->
		  let score' = count_ones cell_size str in
		    if compare_big_int score' score = 1
		    then loop_aux table score' steps' q
		    else loop_aux champion score steps q
	      | MACUNDEF(x,st) ->
		  let q' = gen_tables table maxst nundef x st in
		    loop_aux champion score steps (List.append q' q)
	      | MACCRAZY -> loop_aux champion score steps q
	  end
    in
    let trzero = {write = 1; move = Goleft; new_state = 1} in
    let t = Table.init n in
      Table.add_trans t 0 0 trzero;
      loop_aux t zero_big_int zero_big_int [(t,0,2*n-1)]
	
end
