open Big_int
open Map
open Table
open Turing
open Macro_strip
open Macro_table
open Mini_machine
  
module type TMacro_Machine =
sig
  open Table
  open Macro_Strip
  open Macro_Table

  type tmacroresult =
    | MACSTOP of tstrip * tcount
    | MACUNDEF of tcell * tstate
    | MACCRAZY
    
  val run : ttable -> tsize -> tmacroresult
end

module Macro_Machine =
struct
  open Table
  open Macro_Strip
  open Macro_Table
  open Ordered_Cells
  open Table_Mapping
  open Mini_Machine

  type tmacroresult =
    | MACSTOP of tstrip * tcount
    | MACUNDEF of tcell * tstate
    | MACCRAZY
    
  let one_step ms mt st =
    let x,p = Macro_Strip.read ms in
    let tr = Table_Mapping.find (x,p,st) mt in
      if tr.macro_new_state = st
      then
	let ms',c = app_trans_it ms tr.macro_move tr.macro_write in
	  (ms',mult_big_int c tr.macro_count,st)
      else (app_trans ms tr.macro_move tr.macro_write,tr.macro_count,tr.macro_new_state)

  let run table cellsize =
    let size = Array.length table
    in
    let rec run_aux mstr mtab st steps =
      try
	let mstr',c,st' = one_step mstr mtab st in
	let steps' = add_big_int steps c in
	  if st'=size
	  then MACSTOP(mstr',steps')
	  else
	    if Macro_Strip.length mstr' > 100
	    then MACCRAZY
	    else run_aux mstr' mtab st' (add_big_int steps c)
      with
	| Not_found ->
	    begin
	      let x,p = Macro_Strip.read mstr in
		match Mini_Machine.run table x cellsize p with
		  | MOK(x',p',c,st') ->
		      let new_trans = {macro_write =x';
				       macro_move = p';
				       macro_new_state = st';
				       macro_count = big_int_of_int c} in
		      let mtab' = Table_Mapping.add (x,p,st) new_trans mtab in
			run_aux mstr mtab' st steps
		  | MSTOP(x',c) ->
		      let mstr' = Macro_Strip.write x' mstr
		      and steps' = add_big_int steps (big_int_of_int c) in
			MACSTOP(mstr',steps')
		  | MCRAZY -> MACCRAZY
		  | MUNDEF(x',st') -> MACUNDEF(x',st')
	    end
	| Infinite_Loop -> MACCRAZY
    in
      run_aux (Macro_Strip.init ()) (Table_Mapping.empty) 0 zero_big_int
	
end

    
