open Big_int
open Turing

module type TMacro_Strip = sig
  exception Infinite_Loop

  type tstrip

  val init : unit -> tstrip
  val read : tstrip -> tbigcell * tpos
  val write : tbigcell -> tstrip -> tstrip
  val app_trans : tstrip -> tdep -> tcell -> tstrip
  val app_trans_it : tstrip -> tdep -> tcell -> tstrip * tcount
  val find_one : tstrip -> bool
  val count_ones : tsize -> tstrip -> tcount
  val length : tstrip -> int
end

module Macro_Strip = 
struct
  exception Infinite_Loop

  type tmulticel = tbigcell * tcount
  type tcentralcel = tbigcell * tpos
  type thalfstrip = tmulticel list
  type tstrip = thalfstrip * tcentralcel * thalfstrip

  let init () = [],(0,Right),[]
    
  let read = function _,(x,p),_ -> x,p

  let write x = function g,(_,p),d -> g,(x,p),d

  let get = function
    | [] -> 0,[]
    | (x,one)::q when eq_big_int one unit_big_int -> x,q
    | (x,n)::q -> x,((x,sub_big_int n unit_big_int)::q)
  let add x = function
    | (x',n)::q when x'=x -> (x,add_big_int n unit_big_int)::q
    | l -> (x,unit_big_int)::l
  let left = function g,(x,_),d ->
    let x',g'= get g and d' = add x d in g',(x',Right),d'
  let right = function g,(x,_),d ->
    let x',d'= get d and g' = add x g in g',(x',Left),d'
  let move st = function
    | Goleft -> left st
    | Goright -> right st
  let app_trans st d x = move (write x st) d

  let eat_zero = function
    | [] -> raise Infinite_Loop
    | [(0,n)] -> raise Infinite_Loop
    | (0,n)::q -> n,q
    | l -> zero_big_int,l
  let eat x l =
    if x=0
    then eat_zero l
    else
      match l with
	| (x',n)::q when x'=x -> n,q
	| l -> zero_big_int,l
  let put x n = function
    | (x',n')::q when x'=x -> (x,add_big_int n n')::q
    | l -> (x,n)::l

  let left_it = function g,(x,_),d ->
    let n,g1 = eat x g in
    let x',g2 = get g1 in
    let n' = add_big_int n unit_big_int in
    let d1 = put x n' d in
      (g2,(x',Right),d1),n'
  let right_it = function g,(x,_),d ->
    let n,d1 = eat x d in
    let x',d2 = get d1 in
    let n' = add_big_int n unit_big_int in
    let g1 = put x n' g in
      (g1,(x',Right),d2),n'
  let move_it st = function
    | Goleft -> left_it st
    | Goright -> right_it st
	
  let left_ch_it x = function g,(x',_),d ->
    let n,g1 = eat x' g in
    let x'',g2 = get g1 in
    let n' = add_big_int n unit_big_int in
    let d1 = put x n' d in
      (g2,(x'',Right),d1),n'
  let right_ch_it x = function g,(x',_),d ->
    let n,d1 = eat x' d in
    let x'',d2 = get d1 in
    let n' = add_big_int n unit_big_int in
    let g1 = put x n' g in
      (g1,(x'',Right),d2),n'
  let move_ch_it st x = function
    | Goleft -> left_ch_it x st
    | Goright -> right_ch_it x st
  let app_trans_it st d x =
    match st with
      | _,(x',_),_ when x'=x -> move_it st d
      | _ -> move_ch_it st x d

  let find_one =
    let rec find_one_strip = function
      | [] -> false
      | (x,_)::q -> x <> 0 || find_one_strip q
    in
      function g,(x,_),d ->
	x <> 0 || find_one_strip g || find_one_strip d

  let count_ones size =
    let count_cell x =
      let acc = ref 0
      and x' = ref x
      in
	for i = 0 to size - 1 do
	  acc := !acc + (!x' mod 2);
	  x' := !x' / 2
	done;
	!acc
    in
    let rec count_half_strip acc = function
      | [] -> acc
      | (x,n)::q ->
	  let ccx = big_int_of_int (count_cell x) in
	  let nccx = mult_big_int n ccx in
	  let nccxacc = add_big_int nccx acc in
	    count_half_strip nccxacc q
    in
      function g,(x,_),d ->
	let clstr = (count_half_strip zero_big_int g)
	and cccell = big_int_of_int (count_cell x)
	and crstr = count_half_strip zero_big_int d
	in
	  add_big_int clstr (add_big_int cccell crstr)
      
  let length = function g,_,d ->
    (List.length g) + (List.length d)
end
