open Turing
open Table

module type TMini_Machine =
sig

  open Table
  
  type tminiresult = | MOK of tbigcell * tdep * tminicount * tstate
		 | MSTOP of tbigcell * tminicount
		 | MUNDEF of tcell * tstate
		 | MCRAZY
    
  val run : Table.ttable -> tbigcell -> tsize -> tpos -> tminiresult

end

module Mini_Machine = struct

  open Table

  type tindex = int
  type tminiresult = | MOK of tbigcell * tdep * tminicount * tstate
		 | MSTOP of tbigcell * tminicount
		 | MUNDEF of tcell * tstate
		 | MCRAZY

  exception HLT
  exception UT
    
  type mini_strip = tcell array

  let maxsteps = 1000

  let init_mini_strip x size =
    let t = Array.create size 0
    and x' = ref x
    in
      for i = 0 to size - 1 do
	t.(i) <- !x' mod 2;
	x' := !x' / 2
      done;
      t

  let ministrip_to_bigcell s size =
    let u = ref 0 in
      for i = size - 1 downto 0 do
	u := !u * 2 + s.(i)
      done;
      !u
	
  let run table bigcell size pos =
    let strip = init_mini_strip bigcell size
    and one_step strip index state =
      match table.(strip.(index)).(state) with
	| TRANS(trans) ->
	    strip.(index) <- trans.write;
	    let index' = if trans.move = Goleft then index+1 else index-1 in
	      index',trans.new_state
	| HALT -> raise HLT
	| UNDEF_TRANS -> raise UT
    in
    let rec run_aux steps state index =
      try
	let index',newstate = one_step strip index state in
	  begin
	    match (index'<0,index'=size,steps>maxsteps) with
	      | true,_,_ ->
		  MOK((ministrip_to_bigcell strip size),Goright,(steps+1),newstate)
	      | _,true,_ ->
		  MOK((ministrip_to_bigcell strip size),Goleft,(steps+1),newstate)
	      | _,_,true -> MCRAZY
	      | _,_,_ -> run_aux (steps+1) newstate index'
	  end
      with
	| HLT -> MSTOP((ministrip_to_bigcell strip size),(steps+1))
	| UT -> MUNDEF(strip.(index),state)
    in
    let index = if pos = Left then size-1 else 0
    in
      run_aux 0 0 index
	  
end
