(************************************)
(* File table.ml : transition table *)
(************************************)

open Turing

(* Module Table : transition table of the machines *)
module type TTable = sig
  
  type ttrans = {write : tcell; move : tdep; new_state : tstate}
  type ttransition = | TRANS of ttrans | UNDEF_TRANS | HALT
  type ttable = ttransition array array

  val init : tsize -> ttable
  val copy : ttable -> ttable
  val add_trans : ttable -> tcell -> tstate -> ttrans -> unit
  val add_halt : ttable -> tcell -> tstate -> unit
end

module Table = struct
  
  type ttrans = {write : tcell; move : tdep; new_state : tstate}
  type ttransition = | TRANS of ttrans | UNDEF_TRANS | HALT
  type ttable = ttransition array array

  let init n = Array.make_matrix 2 n UNDEF_TRANS

  let copy t =
    let n = Array.length t in
    let t' = init n in
      for i = 0 to 1 do
	for j = 0 to n-1 do
	  match t.(i).(j) with
	    | TRANS(tr) ->
		t'.(i).(j) <- TRANS({write = tr.write;
				     move = tr.move;
				     new_state = tr.new_state})
	    | HALT -> t'.(i).(j) <- HALT
	    | _ -> ()
	done
      done;
      t'

  let add_trans table read state trans = table.(read).(state) <- TRANS(trans)
					   
  let add_halt table read state = table.(read).(state) <- HALT
end
